"""
Module to create an isolated environment. Like Pipx but allows
the ability to create an environment in a specific directory and
then populate it with packages incrementally. Also allows the
ability to run commands in the environment.
"""

import json
import os
import shutil
import subprocess
import sys
import venv
import warnings
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Iterator

from filelock import FileLock

from isolated_environment.requirements import Requirements

WARN_OPERATOR = [  # _todo: implement operator support
    "==",
    ">=",
    "<=",
    ">",
    "<",
    "~=",
    "!=",
]


def _create_virtual_env(env_path: Path) -> Path:
    """Creates an empty virtual environment in the current directory using venv."""

    # Create parent directories
    env_path.parent.mkdir(parents=True, exist_ok=True)
    # Create the virtual environment
    venv.create(env_path, with_pip=True)
    env_name = env_path.name
    print(f"Virtual environment '{env_name}' created at {env_path}")
    return env_path


def _get_activated_environment(env_path: Path) -> dict[str, str]:
    """Gets the activate environment for the environment."""
    out_env = os.environ.copy()
    if sys.platform == "win32":
        out_env["PATH"] = str(env_path / "Scripts") + ";" + out_env["PATH"]
    else:
        out_env["PATH"] = str(env_path / "bin") + ":" + out_env["PATH"]

    # set PYTHONPATH to make sure Python finds installed packages
    if sys.platform == "win32":
        out_env["PYTHONPATH"] = str(env_path / "Lib" / "site-packages")
    else:
        out_env["PYTHONPATH"] = str(env_path / "lib" / "site-packages")
    return out_env


def _pip_install(env_path: Path, package: str, extra_index: str | None = None) -> None:
    """Installs a package in the virtual environment."""
    # Activate the environment and install packages
    env = _get_activated_environment(env_path)
    cmd_list = ["pip", "install", package]
    if extra_index:
        cmd_list.extend(["--extra-index-url", extra_index])
    cmd = subprocess.list2cmdline(cmd_list)
    print(f"Running: {cmd}")
    subprocess.run(cmd, env=env, shell=True, check=True)


def _package_name(package: str) -> str:
    """Splits a package into name and version."""
    for f in WARN_OPERATOR:
        if f in package:
            return package.split(f)[0]
    return package


class IsolatedEnvironment:
    """An isolated environment."""

    def __init__(
        self, env_path: Path, requirements: Requirements | None = None
    ) -> None:
        self.env_path = env_path
        self.env_path.mkdir(parents=True, exist_ok=True)
        # file_lock is side-by-side with the environment.
        self.file_lock = FileLock(str(env_path) + ".lock")
        self.packages_json = env_path / "packages.json"
        if requirements is not None:
            self.ensure_installed(requirements)

    def install_environment(self) -> None:
        """Installs the environment."""
        assert (
            not self.installed()
        ), f"The environment {self.env_path} is already installed."
        self.env_path = _create_virtual_env(self.env_path)
        # write the packages.json file
        empty_reqs = Requirements([])
        self._write_reqs(empty_reqs)

    def installed(self) -> bool:
        """Returns True if the environment is installed."""
        if not self.env_path.exists():
            return False
        if sys.platform == "win32":
            return self.env_path.exists() and (self.env_path / "Scripts").exists()
        return self.env_path.exists() and (self.env_path / "bin").exists()

    def clean(self) -> None:
        """Cleans the environment."""
        if self.env_path.exists():
            shutil.rmtree(self.env_path, ignore_errors=True)

    def _read_reqs(self) -> Requirements:
        """Reads the packages.json file."""
        if not self.packages_json.exists():
            return Requirements([])
        out = self.packages_json.read_text()
        data = Requirements.from_json(out)
        return data

    def _write_reqs(self, reqs: Requirements) -> None:
        """Writes the packages.json file."""
        out = reqs.to_json()
        self.packages_json.write_text(out)

    @contextmanager
    def lock(self) -> Iterator[None]:
        """Locks the environment to prevent it from being used."""
        self.file_lock.acquire()
        try:
            yield
        finally:
            self.file_lock.release()

    def pip_install(
        self, package: str | list[str], extra_index: str | None = None
    ) -> None:
        """Installs a package in the virtual environment."""
        assert (
            self.env_path.exists()
        ), f"The environment {self.env_path} doesn't exist, install it first."
        reqs = self._read_reqs()
        if isinstance(package, list):
            for p in package:
                _pip_install(self.env_path, p, extra_index)
                reqs.add(package)
        elif isinstance(package, str):
            _pip_install(self.env_path, package, extra_index)
            reqs.add(package)
        else:
            raise TypeError(f"Unknown type for package: {type(package)}")
        self._write_reqs(reqs)

    def environment(self) -> dict[str, str]:
        """Gets the activated environment, which should be applied to subprocess environments."""
        return _get_activated_environment(self.env_path)

    def run(self, cmd_list: list[str]) -> subprocess.CompletedProcess:
        """Runs a command in the environment."""
        env = self.environment()
        cmd = subprocess.list2cmdline(cmd_list)
        cp = subprocess.run(cmd, env=env, shell=True, check=True)
        return cp

    def pip_list(self) -> dict[str, Any]:
        """Returns a dictionary of installed packages."""
        if not self.installed():
            return {}
        env = self.environment()
        cmd_list = ["pip", "list", "--format", "json"]
        cmd = subprocess.list2cmdline(cmd_list)
        completed = subprocess.run(
            cmd,
            env=env,
            shell=True,
            check=True,
            capture_output=True,
            universal_newlines=True,
        )
        stdout = completed.stdout
        out = json.loads(stdout)
        return out  # type: ignore

    def pip_has(self, packages: list[str]) -> bool:
        """Returns True if the packages are installed."""
        pip_list = self.pip_list()
        pip_list_names = [p["name"] for p in pip_list]  # type: ignore
        for p in packages:
            for f in WARN_OPERATOR:
                if f in p:
                    warnings.warn(
                        f"Warning: The package {p} has an operator {f} in "
                        + "it and we don't support version matching yet."
                    )
        for package in packages:
            package_name = _package_name(package)
            if package_name not in pip_list_names:
                return False
        return True

    # Returns an environment dictionary
    def ensure_installed(self, reqs: Requirements) -> dict[str, Any]:
        """Ensures that the packages are installed."""
        with self.lock():
            if not self.installed():
                self.install_environment()
            prev_reqs = self._read_reqs()
            if reqs == prev_reqs:
                return self.environment()
            list_reqs = list(reqs)
            for req in list_reqs:
                if req not in prev_reqs:
                    package_str = str(req)
                    self.pip_install(package_str)
            self._write_reqs(reqs)
            return self.environment()

    def installed_requirements(self) -> Requirements:
        """Returns a list of installed requirements."""
        return self._read_reqs()



"""
Unit test file.
"""

import json
import subprocess
import unittest
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any

from isolated_environment.api import IsolatedEnvironment
from isolated_environment.requirements import Requirements

HERE = Path(__file__).parent
TEST_DIR = HERE / "test"


def pretty(data: Any) -> str:
    """Make JSON beautiful."""
    return json.dumps(data, indent=4, sort_keys=True)


class IsolatedEnvironmentTest(unittest.TestCase):
    """Main tester class."""

    def test_ensure_installed(self) -> None:
        """Tests that ensure_installed works."""
        reqs = Requirements(
            [
                "static-ffmpeg",
            ]
        )
        with TemporaryDirectory() as tmp_dir:
            iso_env = IsolatedEnvironment(Path(tmp_dir) / "venv")
            env = iso_env.ensure_installed(reqs)
            self.assertTrue(iso_env.installed())
            installed_reqs = iso_env.installed_requirements()
            self.assertEqual(installed_reqs, reqs)
            subprocess.check_output(["static_ffmpeg", "--help"], env=env, shell=True)
            # Second time should be a no-op.
            iso_env.ensure_installed(reqs)
            iso_env.clean()
            self.assertFalse(iso_env.installed())


if __name__ == "__main__":
    unittest.main()



"""
Unit test file.
"""
import os
import unittest

COMMAND = "isolated_environment"


class MainTester(unittest.TestCase):
    """Main tester class."""

    @unittest.skip("No CLI yet.")
    def test_imports(self) -> None:
        """Test command line interface (CLI)."""
        rtn = os.system(COMMAND)
        self.assertEqual(0, rtn)


if __name__ == "__main__":
    unittest.main()



import json
from dataclasses import dataclass
from enum import Enum
from typing import Any, Iterator, List, Tuple, Union

import semver


class Operator(Enum):
    """Enum for requirement operator"""

    EQ = "=="
    GE = ">="
    LE = "<="
    GT = ">"
    LT = "<"
    NE = "!="
    APPROX = "~="

    # str representation of the enum
    def __str__(self) -> str:
        return self.value

    # equals operator
    def __eq__(self, other: object) -> bool:
        if isinstance(other, Operator):
            return self.value == other.value
        if isinstance(other, str):
            return self.value == other
        return NotImplemented


def comp(a: Any, b: Any) -> bool:
    """Compares two objects."""
    if a is None and b is None:
        return True
    if a is None or b is None:
        return False
    return a == b


@dataclass
class ParsedReq:
    """Parsed requirement"""

    package_name: str
    enum_operator: Operator | None = None
    semversion: semver.VersionInfo | None = None
    extra_index_url: str | None = None

    @staticmethod
    def parse(requirement: str) -> "ParsedReq":
        # Split the requirement string into package name, version and extra_index_url
        operator = None
        semversion = None
        extra_index_url = None
        if "--extra-index-url" in requirement:
            requirement, extra_index_url = requirement.split("--extra-index-url")
        # strip
        requirement = requirement.strip()
        if extra_index_url is not None:
            extra_index_url = extra_index_url.strip()
        for op in Operator:
            if op.value in requirement:
                package_name, version = requirement.split(op.value)
                operator = op
                semversion = semver.VersionInfo.parse(version)
                break
        else:
            package_name = requirement
        # Return a new ParsedRequirement instance
        return ParsedReq(package_name, operator, semversion, extra_index_url)

    def compare(
        self,
        other_package_name: str,
        other_semversion: semver.VersionInfo | None = None,
        other_extra_index_url: str | None = None,
    ) -> Tuple[bool, bool, bool]:
        """Compares the parsed requirement with another parsed requirement"""
        package_name_matches = comp(self.package_name, other_package_name)
        if not package_name_matches:
            return False, False, False
        semversion_matches = comp(self.semversion, other_semversion)
        extra_index_url_matches = comp(self.extra_index_url, other_extra_index_url)
        return True, semversion_matches, extra_index_url_matches

    def __str__(self) -> str:
        out = []
        out.append(self.package_name)
        if self.enum_operator is not None:
            out.append(str(self.enum_operator))
        if self.semversion is not None:
            out.append(str(self.semversion))
        if self.extra_index_url is not None:
            out.append(f"--extra-index-url {self.extra_index_url}")
        return " ".join(out)


@dataclass
class ParsedReqs:
    """Parsed requirements"""

    reqs: List[ParsedReq]

    def _has_pckg_str(self, line: str) -> bool:
        req = ParsedReq.parse(line)
        return all(
            r.compare(req.package_name, req.semversion, req.extra_index_url)
            for r in self.reqs
        )

    def _has_pckg_req(self, req: ParsedReq) -> bool:
        for r in self.reqs:
            matches = r.compare(req.package_name, req.semversion, req.extra_index_url)
            if matches[0]:
                return all(matches)
            return False
        return False

    def has(self, other: Union["ParsedReqs", ParsedReq, str, List[str]]) -> bool:
        if isinstance(other, str):
            return self._has_pckg_str(other)
        if isinstance(other, ParsedReq):
            return self._has_pckg_req(other)
        if isinstance(other, list):
            return all(self._has_pckg_str(line) for line in other)
        for req in other.reqs:
            if not self._has_pckg_req(req):
                return False
        return True

    def __contains__(self, other: Union["ParsedReqs", str, List[str]]) -> bool:
        return self.has(other)

    def __len__(self) -> int:
        return len(self.reqs)

    def __getitem__(self, index: int) -> ParsedReq:
        return self.reqs[index]

    def __iter__(self) -> Iterator[Any]:
        return iter(self.reqs)


class Requirements:
    """Requirements"""

    def __init__(self, packages: list[str]) -> None:
        assert isinstance(packages, list)
        self._packages = packages
        self._parsed = self._parse()

    def has(self, other: Union["Requirements", ParsedReq, str, List[str]]) -> bool:
        if isinstance(other, Requirements):
            return self._parsed.has(other._parsed)  # pylint: disable=protected-access
        return self._parsed.has(other)

    def add(self, other: Union[str, List[str]]) -> None:
        if isinstance(other, str):
            self._packages.append(other)
        elif isinstance(other, list):
            self._packages.extend(other)
        self._parsed = self._parse()

    def _parse(self) -> ParsedReqs:
        """Parses the requirements"""
        reqs = [ParsedReq.parse(package) for package in self._packages]
        return ParsedReqs(reqs)

    def __contains__(
        self, other: Union["Requirements", ParsedReq, str, List[str]]
    ) -> bool:
        return self.has(other)

    def __len__(self) -> int:
        return len(self._parsed)

    def __getitem__(self, index: int) -> ParsedReq:
        return self._parsed[index]

    def __iter__(self) -> Iterator[Any]:
        return iter(self._parsed)

    def to_json(self) -> str:
        """Returns a JSON representation of the requirements."""
        out = json.dumps(self._packages, indent=4, sort_keys=True)
        return out

    def __str__(self) -> str:
        return self.to_json()

    def __repr__(self) -> str:
        return self.to_json()

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, Requirements):
            return self._packages == __value._packages
        return NotImplemented

    @staticmethod
    def from_json(json_str: str) -> "Requirements":
        """Returns a Requirements instance from a JSON string."""
        return Requirements(json.loads(json_str))



"""
Unit test file.
"""
import unittest

from isolated_environment.requirements import Requirements


class RequirementsTesting(unittest.TestCase):
    """Main tester class."""

    def test_simple_requirements(self) -> None:
        """Tests simple requirements."""
        reqs = Requirements(["package1", "package2", "package3"])

        for req in reqs:
            self.assertIn(req.package_name, ["package1", "package2", "package3"])
            self.assertIsNone(req.enum_operator)
            self.assertIsNone(req.semversion)

    def test_single_package_with_version(self) -> None:
        """Tests single package with version."""
        reqs = Requirements(["package==1.0.0", "package2>=1.0.0"])

        self.assertEqual(len(reqs), 2)
        self.assertEqual(reqs[0].package_name, "package")
        self.assertEqual(reqs[0].enum_operator, "==")
        self.assertEqual(reqs[0].semversion, "1.0.0")

    def test_package_with_greater_less_not_equal_operators(self) -> None:
        """Tests package with greater than, less than, and not equal to operators."""
        reqs = Requirements(["package1>1.0.0", "package2<2.0.0", "package3!=3.0.0"])

        self.assertEqual(len(reqs), 3)
        self.assertEqual(reqs[0].package_name, "package1")
        self.assertEqual(reqs[0].enum_operator, ">")
        self.assertEqual(reqs[0].semversion, "1.0.0")
        self.assertEqual(reqs[1].package_name, "package2")
        self.assertEqual(reqs[1].enum_operator, "<")
        self.assertEqual(reqs[1].semversion, "2.0.0")
        self.assertEqual(reqs[2].package_name, "package3")
        self.assertEqual(reqs[2].enum_operator, "!=")
        self.assertEqual(reqs[2].semversion, "3.0.0")

    def test_package_with_greater_less_equal_operators(self) -> None:
        """Tests package with greater than or equal to, and less than or equal to operators."""
        reqs = Requirements(["package1>=1.0.0", "package2<=2.0.0"])

        self.assertEqual(len(reqs), 2)
        self.assertEqual(reqs[0].package_name, "package1")
        self.assertEqual(reqs[0].enum_operator, ">=")
        self.assertEqual(reqs[0].semversion, "1.0.0")
        self.assertEqual(reqs[1].package_name, "package2")
        self.assertEqual(reqs[1].enum_operator, "<=")
        self.assertEqual(reqs[1].semversion, "2.0.0")

    def test_extra_index_url_mismatch(self) -> None:
        """Tests extra index url mismatch."""
        reqs = Requirements(
            [
                "package1==1.0.0 --extra-index-url https://pypi.org/simple",
                "package2>=1.0.0 --extra-index-url https://test.pypi.org/simple",
            ]
        )

        self.assertEqual(len(reqs), 2)
        self.assertEqual(reqs[0].package_name, "package1")
        self.assertEqual(reqs[0].enum_operator, "==")
        self.assertEqual(reqs[0].semversion, "1.0.0")
        self.assertEqual(reqs[0].extra_index_url, "https://pypi.org/simple")
        self.assertEqual(reqs[1].package_name, "package2")
        self.assertEqual(reqs[1].enum_operator, ">=")
        self.assertEqual(reqs[1].semversion, "1.0.0")
        self.assertEqual(reqs[1].extra_index_url, "https://test.pypi.org/simple")

    def test_extra_index_url_absence(self) -> None:
        """Tests absence of extra index url."""
        reqs = Requirements(
            [
                "package1==1.0.0 --extra-index-url https://pypi.org/simple",
                "package2>=1.0.0",
            ]
        )

        self.assertEqual(len(reqs), 2)
        self.assertEqual(reqs[0].package_name, "package1")
        self.assertEqual(reqs[0].enum_operator, "==")
        self.assertEqual(reqs[0].semversion, "1.0.0")
        self.assertEqual(reqs[0].extra_index_url, "https://pypi.org/simple")
        self.assertEqual(reqs[1].package_name, "package2")
        self.assertEqual(reqs[1].enum_operator, ">=")
        self.assertEqual(reqs[1].semversion, "1.0.0")
        self.assertIsNone(reqs[1].extra_index_url)

    def test_has(self) -> None:
        """Tests absence of extra index url."""
        deps = [
            "package1==1.0.0 --extra-index-url https://pypi.org/simple",
            "package2>=1.0.0",
        ]
        reqs = Requirements(deps)
        self.assertIn("package1==1.0.0 --extra-index-url https://pypi.org/simple", reqs)
        self.assertIn("package2>=1.0.0", reqs)
        self.assertIn(deps, reqs)


if __name__ == "__main__":
    unittest.main()



from pathlib import Path
from subprocess import CompletedProcess
from typing import Any

from .api import IsolatedEnvironment  # noqa: F401
from .requirements import Requirements  # noqa: F401

__all__ = ["isolated_environment"]


def isolated_environment(
    env_path: Path, requirements: list[str] | None = None
) -> dict[str, Any]:
    """Creates an isolated environment."""
    requirements = requirements or []
    reqs = Requirements(requirements)
    iso_env = IsolatedEnvironment(env_path, reqs)
    env = iso_env.environment()
    return env


def isolated_environment_cmd(
    env_path: Path,
    cmd_list: list[str],
    requirements: list[str] | None,
) -> CompletedProcess:
    """Creates an isolated environment."""
    requirements = requirements or []
    reqs = Requirements(requirements)
    iso_env = IsolatedEnvironment(env_path, reqs)
    iso_env.ensure_installed(reqs)
    cp = iso_env.run(cmd_list)
    return cp
